/*
 * Pound - the reverse-proxy load-balancer
 * Copyright (C) 2002 Apsis GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Contact information:
 * Apsis GmbH
 * P.O.Box
 * 8707 Uetikon am See
 * Switzerland
 * Tel: +41-1-920 4904
 * EMail: roseg@apsis.ch
 */

static char *rcs_id = "$Id: pound.c,v 0.9 2002/08/19 08:19:53 roseg Exp roseg $";

/*
 * $Log: pound.c,v $
 * Revision 0.9  2002/08/19 08:19:53  roseg
 * Added support for listening on multiple addresses/ports
 * Added support/configuration for WebDAV (LOCK/UNLOCK)
 * Added support for old-style HTTP/1.0 responses (content to EOF)
 * Fixed threads stack size problem on *BSD (#ifdef NEED_STACK)
 * Fixed problem in URL extraction
 *
 * Revision 0.8  2002/08/01 13:29:15  roseg
 * fixed bug in server timeout/close detection
 * fixed problem with SSL multi-threading
 * header collection
 * extended request patterns as per RFC
 * fixed problem with HEAD response (ignore content length)
 *
 * Revision 0.7  2002/07/23 03:11:27  roseg
 * Moved entirely to BIO (rather then the old comm_)
 * Added HTTPS-specific headers
 * Fixed a few minor problems in the pattern matching
 *
 * Revision 0.6  2002/07/16 21:14:01  roseg
 * added URL groups and matching
 * extended URL reuest matching
 * moved to "modern" regex
 *
 * Revision 0.5  2002/07/04 12:19:13  roseg
 * added config file option
 * split program into multiple files
 * cleaned some code
 *
 * Revision 0.4  2002/06/25 23:19:40  roseg
 * added backend dead/resurect code
 * added HTTP/1.1 chunked transfer mode
 * added timeouts for clients
 * code restructuring
 *
 * Revision 0.3  2002/06/18 13:38:20  roseg
 * Added HTTP/1.1 handling - much faster
 * Fixed problem with SIGPIPE
 *
 * Revision 0.2  2002/06/03 01:00:53  roseg
 * added HTTP/HTTPS checking
 * added logging
 * fixed a few minor errors
 *
 * Revision 0.1  2002/05/31 15:53:14  roseg
 * Initial release
 *
 */

#include    "pound.h"

/* common variables */
int     clnt_to;            /* client timeout */
int     log_level;          /* logging mode - 0, 1, 2 */
int     https_headers;      /* add HTTPS-specific headers */
int     allow_xtd;          /* allow extended HTTP - PUT, DELETE */
int     allow_dav;          /* allow WebDAV - LOCK, UNLOCK */
int     alive_to;           /* allow extended HTTP - PUT, DELETE */
char    **http,             /* HTTP port to listen on */
        **https,            /* HTTPS port to listen on */
        **cert,             /* certificate file */
        *user,              /* user to run as */
        *group,             /* group to run as */
        *root;              /* directory to chroot to */

GROUP   **groups;           /* addresses of possible back-end servers */

regex_t GET,                /* GET request: GET uri HTTP/1.[01] */
        POST,               /* POST request: POST uri HTTP/1.[01] */
        HEAD,               /* HEAD request: HEAD uri HTTP/1.[01] */
        PUT,                /* PUT request: PUT uri HTTP/1.[01] */
        DELETE,             /* DELETE request: DELETE uri HTTP/1.[01] */
        LOCK,               /* LOCK request: LOCK uri HTTP/1.[01] */
        UNLOCK,             /* UNLOCK request: UNLOCK uri HTTP/1.[01] */
        HEADER,             /* Allowed header */
        CHUNKED,            /* Transfer-encoding: chunked header */
        CONT_LEN,           /* Content-length header */
        CHUNK_HEAD;         /* chunk header line */

/*
 * handle SIGTERM - exit
 */
static void
h_term(int sig)
{
    syslog(LOG_NOTICE, "received SIGTERM - exiting...");
    exit(0);
}

/*
 * handle SIGPIPE - exit thread
 */
static void
h_pipe(int sig)
{
    pthread_exit(NULL);
}

/*
 * Pound: the reverse-proxy/load-balancer
 *
 * Arguments:
 *  -H addr,port        http address and port to listen on (optional)
 *  -S addr,port        https address and port to listen on (optional)
 *  -c file             https certificate file (mandatory if -s given)
 *  -t secs             min. time to keep sessions. Default 300.
 *  -u user             user to run as (setuid) (optional)
 *  -g group            group to run as (setgid) (optional)
 *  -d root             directory to chroot to (optional)
 *  -X                  allow extended HTTP requests (PUT, DELETE)
 *  -L [012]            logging level: 0 - none, 1 - standard, 2 - extended
 *  -a secs             time-interval to check for resurected hosts
 *  -f config_file      configuration file - exclusive of other flags
 *  -h                  show flags description and exit
 *  addr,port,priority  one or more hosts to proxy to
 */

int
main(int argc, char **argv)
{
    pthread_t           thr;
    pthread_attr_t      attr;
    int                 *http_sock, *https_sock, clnt_length, i, max_fd, clnt;
    struct sockaddr_in  h_addr, clnt_addr;
    struct hostent      *host;
    fd_set              socks;
    uid_t               user_id;
    gid_t               group_id;
    FILE                *fpid;
    regex_t             LISTEN_ADDR;
    regmatch_t          matches[3];

    openlog("pound", LOG_CONS, LOG_DAEMON);
    syslog(LOG_NOTICE, "starting...");

    signal(SIGTERM, h_term);
    /*
     * to avoid problems with some non-standard SSL clients, Konqueror in particualr
     */
    signal(SIGPIPE, h_pipe);

    config_parse(argc, argv);

    /* prepare regular expressions */
    if(regcomp(&GET, "^GET ([A-Za-z0-9;/?:%@&=+$,_.!'()-]+) HTTP/1.[01]$",
        REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&POST, "^POST ([A-Za-z0-9;/?:%@&=+$,_.!'()-]+) HTTP/1.[01]$",
        REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&HEAD, "^HEAD ([A-Za-z0-9;/?:%@&=+$,_.!'()-]+) HTTP/1.[01]$",
        REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&PUT, "^PUT ([A-Za-z0-9;/?:%@&=+$,_.!'()-]+) HTTP/1.[01]$",
        REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&DELETE, "^DELETE ([A-Za-z0-9;/?:%@&=+$,_.!'()-]+) HTTP/1.[01]$",
        REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&LOCK, "^LOCK ([A-Za-z0-9;/?:%@&=+$,_.!'()-]+) HTTP/1.[01]$",
        REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&UNLOCK, "^UNLOCK ([A-Za-z0-9;/?:%@&=+$,_.!'()-]+) HTTP/1.[01]$",
        REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&HEADER, "^[A-Za-z][A-Za-z0-9;/?:%@&=+$,_.!'()-]*:.*$",
        REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&CHUNKED, "^Transfer-encoding: chunked$", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&CHUNK_HEAD, "^([0-9a-f]+).*$", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&CONT_LEN, "^Content-length: ([1-9][0-9]*)$", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    || regcomp(&LISTEN_ADDR, "^([^,]+),([1-9][0-9]*)$", REG_ICASE | REG_NEWLINE | REG_EXTENDED)
    ) {
        syslog(LOG_ERR, "bad Regex - aborted");
        exit(1);
    }

    max_fd = 0;

    /* get HTTP address and port */
    if(http[0]) {
        for(i = 0; http[i]; i++)
            ;
        if((http_sock = (int *)malloc(sizeof(int) * i)) == NULL) {
            syslog(LOG_ERR, "http_sock out of memory - aborted");
            exit(1);
        }
        for(i = 0; http[i]; i++) {
            memset(&h_addr, 0, sizeof(h_addr));
            h_addr.sin_family = AF_INET;

            /* host */
            if(regexec(&LISTEN_ADDR, http[i], 3, matches, 0)) {
                syslog(LOG_ERR, "bad HTTP spec %s - aborted", http[i]);
                exit(1);
            }
            http[i][matches[1].rm_eo] = '\0';
            if((host = gethostbyname(http[i])) == NULL) {
                syslog(LOG_ERR, "Unknown HTTP host %s", http[i]);
                exit(1);
            }
            memcpy(&h_addr.sin_addr.s_addr, host->h_addr_list[0], sizeof(h_addr.sin_addr.s_addr));
            /* port */
            h_addr.sin_port = (in_port_t)htons(atoi(http[i] + matches[2].rm_so));

            /* prepare the socket */
            if((http_sock[i] = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
                syslog(LOG_ERR, "HTTP socket create: %m - aborted");
                exit(1);
            }
            if(http_sock[i] > max_fd)
                max_fd = http_sock[i];
            if(bind(http_sock[i], (struct sockaddr *)&h_addr, (socklen_t)sizeof(h_addr)) < 0) {
                syslog(LOG_ERR, "HTTP socket bind: %m - aborted");
                exit(1);
            }
            listen(http_sock[i], 20);
        }
    }

    /* get HTTPS address and port */
    if(https[0]) {
        for(i = 0; https[i]; i++)
            ;
        if((https_sock = (int *)malloc(sizeof(int) * i)) == NULL) {
            syslog(LOG_ERR, "https_sock out of memory - aborted");
            exit(1);
        }
        for(i = 0; https[i]; i++) {
            memset(&h_addr, 0, sizeof(h_addr));
            h_addr.sin_family = AF_INET;

            /* host */
            if(regexec(&LISTEN_ADDR, https[i], 3, matches, 0)) {
                syslog(LOG_ERR, "bad HTTP spec %s - aborted", https[i]);
                exit(1);
            }
            https[i][matches[1].rm_eo] = '\0';
            if((host = gethostbyname(https[i])) == NULL) {
                syslog(LOG_ERR, "Unknown HTTP host %s", https[i]);
                exit(1);
            }
            memcpy(&h_addr.sin_addr.s_addr, host->h_addr_list[0], sizeof(h_addr.sin_addr.s_addr));
            /* port */
            h_addr.sin_port = (in_port_t)htons(atoi(https[i] + matches[2].rm_so));

            /* prepare the socket */
            if((https_sock[i] = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
                syslog(LOG_ERR, "HTTPS socket create: %m - aborted");
                exit(1);
            }
            if(https_sock[i] > max_fd)
                max_fd = https_sock[i];
            if(bind(https_sock[i], (struct sockaddr *)&h_addr, (socklen_t)sizeof(h_addr)) < 0) {
                syslog(LOG_ERR, "HTTPS socket bind: %m - aborted");
                exit(1);
            }
            listen(https_sock[i], 20);
        }
    }

    max_fd++;

    /* thread stuff */
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    /* set uid if necessary */
    if(user) {
        struct passwd   *pw;

        if((pw = getpwnam(user)) == NULL) {
            syslog(LOG_ERR, "no such user %s - aborted", user);
            exit(1);
        }
        user_id = pw->pw_uid;
    }

    /* set gid if necessary */
    if(group) {
        struct group    *gr;

        if((gr = getgrnam(group)) == NULL) {
            syslog(LOG_ERR, "no such group %s - aborted", group);
            exit(1);
        }
        group_id = gr->gr_gid;
    }

#ifdef  AEMON
    /* daemonize - make ourselves a subprocess. */
    switch (fork()) {
        case 0:
            close(0);
            close(1);
            close(2);
            break;
        case -1:
            syslog(LOG_ERR, "fork: %m - aborted");
            exit(1);
        default:
            exit(0);
    }
#endif

    /* record pid in var/run */
    if((fpid = fopen("/var/run/pound.pid", "wt")) != NULL) {
        fprintf(fpid, "%d\n", getpid());
        fclose(fpid);
    } else
        syslog(LOG_WARNING, "/var/run/pound.pid %m");

    /* chroot if necessary */
    if(root) {
        if(chroot(root)) {
            syslog(LOG_ERR, "chroot: %m - aborted");
            exit(1);
        }
        if(chdir("/")) {
            syslog(LOG_ERR, "chroot/chdir: %m - aborted");
            exit(1);
        }
    }

    if(group)
        if(setgid(group_id) || setegid(group_id)) {
            syslog(LOG_ERR, "setgid: %m - aborted");
            exit(1);
        }
    if(user)
        if(setuid(user_id) || seteuid(user_id)) {
            syslog(LOG_ERR, "setuid: %m - aborted");
            exit(1);
        }

    /* start the pruner */
    if(pthread_create(&thr, &attr, thr_prune, NULL)) {
        syslog(LOG_ERR, "create thr_prune: %m - aborted");
        exit(1);
    }

    /* start resurector (if necessary) */
    if(pthread_create(&thr, &attr, thr_resurect, NULL)) {
        syslog(LOG_ERR, "create thr_resurect: %m - aborted");
        exit(1);
    }

#ifdef  NEED_STACK
    /* set new stack size - necessary for OpenBSD/FreeBSD */
    if(pthread_attr_setstacksize(&attr, 1 << 18)) {
        syslog(LOG_ERR, "can't set stack size - aborted");
        exit(1);
    }
#endif

    /* and start working */
    for(;;) {
        FD_ZERO(&socks);
        for(i = 0; http[i]; i++)
            FD_SET(http_sock[i], &socks);
        for(i = 0; https[i]; i++)
            FD_SET(https_sock[i], &socks);
        if((i = select(max_fd, &socks, NULL, NULL, NULL)) < 0) {
            syslog(LOG_WARNING, "select: %m");
        } else {
            for(i = 0; http[i]; i++) {
                if(FD_ISSET(http_sock[i], &socks)) {
                    memset(&clnt_addr, 0, sizeof(clnt_addr));
                    clnt_length = sizeof(clnt_addr);
                    if((clnt = accept(http_sock[i], (struct sockaddr *)&clnt_addr, (socklen_t *)&clnt_length)) < 0) {
                        syslog(LOG_WARNING, "HTTP accept: %m");
                    } else if (clnt_addr.sin_family != AF_INET) {
                        /* may happen on FreeBSD, I am told */
                        syslog(LOG_WARNING, "HTTP connection prematurely closed by peer");
                        close(clnt);
                    } else {
                        thr_arg *arg;

                        if((arg = (thr_arg *)malloc(sizeof(thr_arg))) == NULL) {
                            syslog(LOG_WARNING, "HTTP arg: malloc");
                            close(clnt);
                        } else {
                            arg->sock = clnt;
                            arg->from_host = clnt_addr.sin_addr;
                            arg->is_ssl = 0;
                            arg->cert = NULL;
                            if(pthread_create(&thr, &attr, thr_http, (void *)arg)) {
                                syslog(LOG_WARNING, "HTTP pthread_create: %m");
                                free(arg);
                                close(clnt);
                            }
                        }
                    }
                }
            }
            for(i = 0; https[i]; i++) {
                if(FD_ISSET(https_sock[i], &socks)) {
                    memset(&clnt_addr, 0, sizeof(clnt_addr));
                    clnt_length = sizeof(clnt_addr);
                    if((clnt = accept(https_sock[i], (struct sockaddr *)&clnt_addr, (socklen_t *)&clnt_length)) < 0) {
                        syslog(LOG_WARNING, "HTTPS accept: %m");
                    } else if (clnt_addr.sin_family != AF_INET) {
                        /* may happen on FreeBSD, I am told */
                        syslog(LOG_WARNING, "HTTPS connection prematurely closed by peer");
                        close(clnt);
                    } else {
                        thr_arg *arg;

                        if((arg = (thr_arg *)malloc(sizeof(thr_arg))) == NULL) {
                            syslog(LOG_WARNING, "HTTPS arg: malloc");
                            close(clnt);
                        } else {
                            arg->sock = clnt;
                            arg->from_host = clnt_addr.sin_addr;
                            arg->is_ssl = 1;
                            arg->cert = cert[i];
                            if(pthread_create(&thr, &attr, thr_http, (void *)arg)) {
                                syslog(LOG_WARNING, "HTTPS pthread_create: %m");
                                free(arg);
                                close(clnt);
                            }
                        }
                    }
                }
            }
        }
    }
}
